from __future__ import annotations as __anotations
from collections.abc import Callable as __Callable
import warnings
import glob
import os
import copy
import numpy as np
import scipy.stats as Stats
import scipy.cluster.hierarchy as Hierarchy
import scipy.spatial.distance as Distance
import matplotlib.pyplot as plt


class TDobj:
    def __init__(self, name: str = "No Name", meta: dict = None,
                 data: dict = None, denominator: float = None, datalen: int = 1000):

        self.meta: dict = dict(sorted(meta.items(), key=lambda x: x[0]))
        self.data: dict = dict(sorted(data.items(), key=lambda x: x[0]))
        self.denominator: float = denominator
        self.datalen: int = datalen
        self.name: str = name
        self.trials_count: int = len(meta["DURATION"])

        if meta is None or data is None or denominator is None or self.trials_count != len(meta["DELAY"]):
            raise Exception("Data may be corrupted.")

    def __add__(self, other) -> TDobj:
        if not other:
            return self

        new_meta: dict = copy.deepcopy(self.meta)
        new_data: dict = copy.deepcopy(self.data)
        new_denominator: float = self.denominator

        if isinstance(other, TDobj):
            if {*self.data} ^ {*other.data}:
                raise Exception("Different formats are mixed. Please review attribute words.")

            if self.datalen != other.datalen:
                raise Exception("Data of different lengths are mixed. Please review \"datalen\" field.")

            if {*self.meta} ^ {*other.meta}:
                warnings.warn("Metadata of different types are mixed. Please review \"meta\" field.", Warning)

            for key in {*self.meta} & {*other.meta}:
                new_meta[key] += other.meta[key]

            for key in self.data:
                new_data[key] += other.data[key]

            new_denominator += other.denominator
        else:
            raise TypeError("Unsupported operand type(s)")

        return TDobj(meta=new_meta, data=new_data, denominator=new_denominator, datalen=self.datalen)

    def __radd__(self, other) -> TDobj:
        return self + other

    def __iadd__(self, other) -> TDobj:
        self = self + other
        return self

    def __bool__(self) -> bool:
        return True

    def getAverageDuration(self) -> float:
        return sum(self.meta["DURATION"]) / self.trials_count

    def getAverageDelay(self) -> float:
        return sum(self.meta["NORMALIZED_DELAY"]) / len(self.meta["NORMALIZED_DELAY"])

    def getAverageDD(self) -> dict:
        result = {}
        count = len(self.meta["DOMINANCE_DURATION"])
        for elem in self.meta["DOMINANCE_DURATION"]:
            for key in elem:
                result[key] = result[key] + elem[key] / count if key in result else elem[key] / count
        return result

    def getDelayAsArray(self) -> np.ndarray:
        normalized_delay = self.meta["NORMALIZED_DELAY"]
        data_count = len(normalized_delay)
        return sum(np.array(
            [self.denominator] * int(normalized_delay[i] * self.datalen) +
            [0] * (self.datalen - int(normalized_delay[i] * self.datalen)))
            for i in range(data_count)) / data_count

    def getMeta(self, key: str) -> list:
        return self.meta[key.upper()]

    def distance(self, other: TDobj) -> float:
        if self.datalen != other.datalen:
            raise Exception("Different formats are mixed")

        temp: np.ndarray = (self.getDelayAsArray() / self.denominator - other.getDelayAsArray() / other.denominator) ** 2
        temp += sum((self.data[key] / self.denominator - other.data[key] / other.denominator) ** 2 for key in self.data)

        result: float = np.mean((temp / 2) ** 0.5)
        return result

    # test ###########################
    def distanceTest1(self, other: TDobj) -> float:
        if self.datalen != other.datalen:
            raise Exception("Different formats are mixed")

        temp: np.ndarray = (self.getDelayAsArray() / self.denominator) * (other.getDelayAsArray() / other.denominator)
        temp += sum((self.data[key] / self.denominator) * (other.data[key] / other.denominator) for key in self.data)

        result: float = np.prod(temp ** (1 / self.datalen))
        return result

    def distanceTest2(self, other: TDobj) -> float:
        if self.datalen != other.datalen:
            raise Exception("Different formats are mixed")

        temp: np.ndarray = (self.getDelayAsArray() / self.denominator) * (other.getDelayAsArray() / other.denominator)
        temp += sum((self.data[key] / self.denominator) * (other.data[key] / other.denominator) for key in self.data)

        result: float = np.mean(temp)
        return result

    def generateClosestTrial(self):
        data = copy.deepcopy(self.data)
        delay = self.getDelayAsArray()
        c = 0
        for i in range(self.datalen):
            label = None
            now = delay[i]
            for key in data:
                if data[key][i] > now:
                    label = key
                    now = data[key][i]
                else:
                    data[key][i] = 0
            for key in data:
                if key == label:
                    data[key][i] = 1
                else:
                    data[key][i] = 0
            if label is None:
                c += 1

        return TDobj(meta={"DURATION": [1.0], "DELAY": [c / self.datalen], "NORMALIZED_DELAY": [
                     c / self.datalen]}, data=data, denominator=1.0, datalen=self.datalen)

    # end test ###########################

    def ddDistance(self, other: TDobj) -> float:
        if self.datalen != other.datalen:
            raise Exception("Different formats are mixed")

        selfdd = self.getAverageDD()
        otherdd = other.getAverageDD()
        result: float = (self.getAverageDelay() - other.getAverageDelay()) ** 2

        result += sum((selfdd[key] - otherdd[key]) ** 2 for key in selfdd)
        result = (result / 2) ** 0.5
        return result

    def setDenominator(self, denominator: float = 1.0) -> TDobj:
        for key in self.data:
            self.data[key] *= denominator / self.denominator

        self.denominator = denominator
        return self

    def setName(self, name: str) -> TDobj:
        self.name = name
        return self

    def convolve(self, weight: int) -> TDobj:
        for key in self.data:
            edge = int(weight / 2)
            edged = np.concatenate([[0] * edge, self.data[key], [self.data[key][-1]] * (weight - edge - 1)])
            self.data[key] = np.convolve(edged, [1 / weight] * weight, "valid")

        return self

    def draw(self, layout: list = None, scale: float = 1.0, size: tuple[float] = (12, 8), font: str = "sans-serif",
             font_size: int = None, show: bool = True, auto_adjust: bool = True, show_legend: bool = True,
             show_chance: bool = True, show_signif: bool = True, show_total: bool = False,
             show_average_delay: bool = False) -> TDobj:

        if layout is None:
            layout = [[[*self.data]]]

        if font:
            plt.rcParams["font.family"] = font
        if font_size:
            plt.rcParams["font.size"] = font_size

        rows = len(layout)
        columns = max(len(row) for row in layout)
        axes = [[None] * columns] * rows
        count = 1
        standard_ax = np.arange(0, self.datalen) / self.datalen * scale

        fig = plt.figure(figsize=size)
        fig.suptitle(self.name)

        for i, row in enumerate(layout):
            for j, column in enumerate(row):
                axes[i][j] = (fig.add_subplot(rows, columns, count, xlabel="Normalized Time", ylabel="Dominance Proportion")
                              if auto_adjust else
                              fig.add_subplot(rows, columns, count, xlabel="Normalized Time", ylabel="Dominance Proportion",
                                              ylim=(0, self.denominator)))
                count += 1

                attributes_count = 1
                for key in column:
                    style: str = "dashed" if attributes_count > 10 else "solid"
                    axes[i][j].plot(standard_ax, self.data[key], label=key, linestyle=style)
                    attributes_count += 1

                # graph parameters
                chance = 1 / len(self.data)
                signif = chance + 1.64485362695145 * (chance * (1 - chance) / self.trials_count) ** 0.5
                average_delay = self.getAverageDelay() * scale

                show_chance and axes[i][j].plot((0, scale),
                                                (chance * self.denominator,) * 2,
                                                label="Chance",
                                                linestyle="dotted",
                                                color="gray")
                show_signif and axes[i][j].plot((0, scale),
                                                (signif * self.denominator,) * 2,
                                                label="Signif.",
                                                linestyle="dashed",
                                                color="gray")
                show_total and axes[i][j].plot(standard_ax,
                                               sum(self.data[key] for key in self.data),
                                               label="Total",
                                               linestyle="dashed")

                show_average_delay and axes[i][j].plot((average_delay, average_delay),
                                                       (0, self.denominator),
                                                       label="Average Delay",
                                                       linestyle="dotted")

                show_legend and axes[i][j].legend()

        # show graph or not
        show and plt.show()

        return self

    def save(self, destination: str = ".", name: str = None, delimiter: str = "\t", file_extension: str = ".csv") -> None:
        if not name:
            name = self.name
        with open(F'{destination}/{name}{file_extension}', "w", encoding="UTF-8") as f:
            attrs = self.data.keys()
            result = delimiter.join(attrs) + "\n" +\
                "\n".join(delimiter.join(F'{j}00000'[:6] for j in i)
                          for i in np.array([self.data[attr]
                                             for attr in attrs]).T)
            f.write(result)


class PCAobj:
    def __init__(self, object_array: TDcontainer, number_of_pcs: int = 2,
                 clustering_method: str = "ward", distance_func: __Callable = TDobj.distance,
                 print_result: bool = True):
        print("> calculating...\n")
        self.object_array: TDcontainer = object_array
        self.number_of_pcs: int = number_of_pcs
        self.label: tuple = tuple(object_array[0].data)

        table: list = []
        for individual in object_array:
            elem = individual.meta["DOMINANCE_DURATION"][0]
            data = [None] * len(self.label)
            for key in elem:
                data[self.label.index(key)] = elem[key]

            table.append(data)

        self.data_matrix: np.ndarray = Stats.zscore(table).T

        # clustering
        if isinstance(distance_func, str):
            distance_func = getattr(TDobj, distance_func)
        self.distance_matrix: np.ndarray = Distance.squareform(
            [[distance_func(object_array[i], object_array[j]) for i in range(len(object_array))]
             for j in range(len(object_array))]
        )
        self.linkage: np.ndarray = Hierarchy.linkage(self.distance_matrix, method=clustering_method)
        self.clusters: np.ndarray = Hierarchy.fcluster(self.linkage, t=0.7 * max(self.linkage[:, 2]), criterion='distance')
        self.corrcoef: np.ndarray = np.corrcoef(self.data_matrix)

        # eigenvalues, eigenvectors
        [w, v] = np.linalg.eig(self.corrcoef)
        v: np.ndarray = v.T
        index = np.argsort(w)[::-1]

        self.eigenvalues: np.ndarray = np.array([w[i] for i in index])
        self.eigenvectors: np.ndarray = np.array([v[i] for i in index])

        # pc scores and vectors for each item
        self.pc: np.ndarray = np.array([self.data_matrix.T @ self.eigenvectors[i] for i in range(number_of_pcs)])
        self.vpc: np.ndarray = np.array([np.identity(len(self.eigenvectors.T)) @ self.eigenvectors[i]
                                         for i in range(number_of_pcs)])

        if print_result:
            for i in range(number_of_pcs):
                print(F'Score of PC{i + 1}:\n{self.pc[i]}\n')

            print(F'Eigenvalues:\n{self.eigenvalues}')
            print(F'Each Contribution Rate (%):\n{self.eigenvalues / len(self.eigenvalues) * 100}\n')
            print(F'Eigenvectors:\n{self.eigenvectors}')

    def draw(self, size: tuple[float] = (6, 4), font: str = "sans-serif", point_size: int = 15,
             show: bool = True, show_legend: bool = True, color_cluster: bool = False,
             show_dendrogram: bool = False, show_scatter: bool = True) -> PCAobj:

        if show_scatter:
            for i in range(len(self.pc) - 1):
                for j in range(i + 1, len(self.pc)):
                    [xlabel, ylabel] = [F'PC{i + 1}', F'PC{j + 1}']
                    fig = plt.figure(figsize=size)
                    fig.suptitle(F'{xlabel} and {ylabel}')
                    ax = fig.add_subplot(xlabel=F'{xlabel} ({int(self.eigenvalues[i] / len(self.label) * 100)}%)',
                                         ylabel=F'{ylabel} ({int(self.eigenvalues[j] / len(self.label) * 100)}%)')
                    if color_cluster:
                        c: int = 0
                        color = plt.rcParams["axes.prop_cycle"].by_key()["color"]
                        noncflg = True
                        for k in range(max(self.clusters)):
                            xp = [self.pc[i][m] for m in range(len(self.clusters)) if self.clusters[m] - 1 == k]
                            yp = [self.pc[j][m] for m in range(len(self.clusters)) if self.clusters[m] - 1 == k]
                            cflg = len(xp) > 1
                            if cflg:
                                c += 1
                                ax.scatter(xp, yp, color=color[c % 10], label=F'cluster{c}', s=point_size)
                            elif noncflg:
                                ax.scatter(xp, yp, color=color[0], label="not in cluster", s=point_size)
                                noncflg = False
                            else:
                                ax.scatter(xp, yp, color=color[0], s=point_size)
                    else:
                        ax.scatter(self.pc[i], self.pc[j], s=point_size)

                    for k in range(len(self.label)):
                        style: str = "dashed" if k > 10 else "solid"
                        ax.plot([0, self.vpc[i][k]], [0, self.vpc[j][k]], label=self.label[k], linestyle=style)

                    ax.grid()
                    show_legend and ax.legend()

        if show_dendrogram:
            fig = plt.figure(figsize=size)
            fig.add_subplot()
            Hierarchy.dendrogram(self.linkage)

        show and plt.show()
        return self


class TDcontainer(list[TDobj]):
    def __add__(self, other):
        return TDcontainer(list(self) + list(other))

    def merge(self) -> TDobj:
        return sum(self).setDenominator(1)

    def mergeAs(self, name: str) -> TDobj:
        return self.merge().setName(name)

    def pca(self, number_of_pcs: int = 2, clustering_method: str = "ward",
            distance_func: __Callable = TDobj.distance, print_result: bool = True) -> PCAobj:
        return PCAobj(number_of_pcs=number_of_pcs, object_array=self,
                      clustering_method=clustering_method, distance_func=distance_func, print_result=print_result)

    def distance(self, distance_func: __Callable = TDobj.distance) -> list[float]:
        data_count: int = len(self)
        result: list[float] = [1] * data_count
        merged = self.merge()
        for i in range(data_count):
            result[i] = distance_func(self[i], merged)
        return result

    def boxPlot(self, size: tuple[float] = (6, 4), whis: float = 1.5, point_size: int = 10, x_position: float = 0.99,
                auto_adjust: bool = True, show: bool = True) -> TDcontainer:
        fig = plt.figure(figsize=size)
        fig.suptitle("Box Plot")
        ax = fig.add_subplot(ylabel="Distance")
        ax.boxplot(self.distance(), sym="", whis=whis)
        auto_adjust or ax.set_ylim(0, 1)
        ax.scatter((x_position,) * len(self), self.distance(), s=point_size)
        show and plt.show()

        return self


def getDataFromStr(tab_separated_values: str, denominator: float = 1.0, datalen: int = 1000) -> TDobj:
    data: dict[str, np.ndarray] = {}
    meta: dict[str, list] = {}
    time: dict[str, list[float]] = {}
    lines: list[str] = tab_separated_values.replace("\r\n", "\n").replace("\r", "\n").split("\n")

    for line in lines:
        if line in {""} or line[0] in {":", "\t"}:
            continue
        params: list[str] = line.split("\t")
        paramlen: int = len(params)
        if paramlen > 1:
            if line[0] == "#":
                tag: str = params[0][1:].strip().upper()
                meta[tag] = [float(param) for param in params[1:]] if tag in {"DURATION", "DELAY"} else params[1:]
            else:
                time[params[0].strip()] = [float(param) for param in params[1:] if param != ""]

        elif paramlen == 1 and params[0] != "":
            if line[0] == "#":
                meta[params[0][1:].strip().upper()] = []
            else:
                time[params[0].strip()] = []

    if "DURATION" not in meta:
        meta["DURATION"] = [max(max(time[key]) for key in time)]
        print("> The duration data was not found, so it was automatically generated")

    if "DELAY" not in meta:
        meta["DELAY"] = [min(min(time[key]) for key in time)]
        print("> The delay data was not found, so it was automatically generated")

    duration: float = meta["DURATION"][0]
    delay: float = meta["DELAY"][0]
    dominance_durs: dict[str, float] = {}

    for key in time:
        sample: list = time[key]
        data[key] = np.zeros(datalen, float)
        dominance_duration: float = 0.0
        for i in range(0, len(sample), 2):
            timebegin = sample[i]
            timeend = sample[i + 1]
            dominance_duration += timeend - timebegin
            data[key][int(timebegin / duration * datalen): int(timeend / duration * datalen)] = denominator

        dominance_durs[key] = dominance_duration / duration

    meta["NORMALIZED_DELAY"] = [delay / duration]
    meta["DOMINANCE_DURATION"] = [dominance_durs]

    return TDobj(meta=meta, data=data, denominator=denominator, datalen=datalen)


def loadFile(file_path: str, denominator: float = 1.0, datalen: int = 1000) -> TDobj:
    """This function generates a TDobj object from a single file.
    The file must be in the TDsampler output format.

    ## Args
    * file_path (str): File path including the extension.
    * denominator (float): Dominance Proportion will be expressed as a fraction of this parameter. Defaults to 1.0.
    * datalen (int): Number of discretization of the whole time (start to stop). Defaults to 1000.

    ## Returns
    * TDobj: An object that contains the time-series data for each attribute word. \
    Each time-series data is represented as a np.ndarray of the size of "denominator" parameter. \
    And its "name" member will be set to the file name (excluding the file extension).
    """
    file_path: str = file_path.replace("\\", "/")
    try:
        with open(file_path, "r", encoding="UTF-8") as tdDataFile:
            print(F'> loading "{file_path}"...')
            return (getDataFromStr(tdDataFile.read(), denominator=denominator, datalen=datalen)
                    .setName(os.path.splitext(os.path.basename(file_path))[0]))
    except Exception as e:
        print(e)
    return None


def loadFolder(folder_path: str = ".", file_extension: str = ".csv",
               denominator: float = 1.0, datalen: int = 1000) -> TDcontainer:
    """This function generates a TDcontainer object from files in the specified folder(s).
    All files must be in the TDsampler output format.

    ## Args
    * folder_path (str): Folder path (not the file path). You can use regular expressions. Defaults to ".".
    * file_extension (str): File extension. Defaults to ".csv".
    * denominator (float): Dominance Proportion will be expressed as a fraction of this parameter. Defaults to 1.0.
    * datalen (int): Number of discretization of the whole time (start to stop). Defaults to 1000.

    ## Returns
    * TDcontainer: A list-like object that contains multiple TDobj objects.
    """
    folder_path: str = folder_path.replace("\\", "/")
    print(F'> loading "{folder_path}"...')

    files: list = glob.glob(F'{folder_path}/*{file_extension}')
    TDobj_list: TDcontainer = TDcontainer(loadFile(f, denominator=denominator, datalen=datalen) for f in files)

    if not len(TDobj_list):
        raise Exception("Folder seems to be empty")

    return TDobj_list


plt.rcParams["font.sans-serif"][0:0] = ("Meiryo", "Yu Gothic", "MS Gothic", "Hiragino Sans")
