from typing import Callable as __Callable, Iterable as __Iterable
import PySimpleGUI as __sg
import os as __os
import time as __time
import datetime as __datetime
import pprint as __pprint
import random as __random


# OPTIONS
attribute_list_path: str = "./attribute.txt"             # file path to attribute list (attribute.txt)
attributes: __Iterable[str] | str = None                 # instead of attribute.txt, this can be used
assessor_name: str = "unknown"                           # assessor name
custom_metadata: dict[str, __Iterable] = {}              # custom metadata
comments: str | __Iterable[str] = None                   # comments
shuffle: bool = True                                     # attribute buttons will be shuffled if this is True
button_text_color: str = "#ffffff"                       # text color of buttons
button_color_on: str = "#008800"                         # background color of buttons (on)
button_color_off: str = "#082567"                        # background color of buttons (off)
button_color_disabled: str = "#666666"                   # background color of buttons (disabled)
button_color_start: str = "#bb0000"                      # background color of start button
button_color_stop: str = "#bb0000"                       # background color of stop button
button_justification: str = "center"                     # button justification (center, right, left)
button_padding: int = 20                                 # button padding
output_folder: str = "output"                            # output folder name
output_file_prefix: str = "out"                          # output file prefix
output_file_joint: str = "-"                             # output file joint
output_file_number: int | str | list[str] = 0            # output file number
output_file_increment: bool = True                       # auto increment file number if this is True
output_file_suffix: str = ""                             # output file suffix
output_file_extension: str = ".csv"                      # output file extension
after_task_scoring: tuple[int] = None                    # range of scoring
max_column: int = 3                                      # maximum count of button columns
font_family: str = "Meiryo UI"                           # font family for gui components
font_size: int = 14                                      # font size for gui components
title_font_size: int = 25                                # font size of title
theme: str = "SystemDefault"                             # PySimpleGUI theme


# MAIN
def run() -> None:
    # print status
    print("\n> Constructing GUI...")

    # set PySimpleGUI theme
    __sg.theme(theme)

    # variables for gui components
    buttons: list[str]             # button labels
    buttonField: list[list] = [[]]  # button components
    maxTextLen: int = 0             # use as button size

    commonFrameArgs: dict = {"title": "", "border_width": 0}
    commonButtonArgs: dict = {"auto_size_button": True, "focus": False, "bind_return_key": False}

    # create button labels from attribute words
    if attributes is None:
        __newFile(attribute_list_path, "button1\nbutton2\nbutton3")  # make a new txt file of atribute words if not exist
        try:
            with open(attribute_list_path, "r", encoding="UTF-8") as attrs:
                buttons = __attributetxt2list(attrs.read(), shuffle=shuffle)
        except Exception as e:
            print(e)
    else:
        buttons = __attributetxt2list(attributes if isinstance(attributes, str) else "\n".join(attributes), shuffle=shuffle)

    # create button components from button labels
    for elem in buttons:
        elemWidth: int = sum(1 + (len(str.encode(char)) >= 2) for char in elem)
        lastElem: list = buttonField[-1]
        buttonSize: int = 10 + (elemWidth >= 10)
        maxTextLen: int = max(elemWidth, maxTextLen)

        lastElem.append(__sg.Submit(
            **commonButtonArgs,
            button_text=elem,
            button_color=(button_text_color, button_color_disabled),
            size=(buttonSize, None),
            pad=((button_padding, button_padding), (0, button_padding * 2)),))

        len(lastElem) == max_column and buttonField.append([])

    # gui components
    startButton: list[list] = [
        [__sg.Submit(
            **commonButtonArgs,
            button_text=startButtonList,
            button_color=button_color_start,
            pad=((button_padding, button_padding), (button_padding, button_padding * 2)),
        ) for startButtonList in {" START ", " STOP ", " SAVE & RESET "}]
    ]
    status: list[list] = [
        [__sg.Text(
            text="Not Started ⏹",
            text_color="#000000",
            key=" status ")]
    ]
    outputFolderSetting: list[list] = [
        [__sg.Text("Output Folder:")],
        [__sg.Input(
            default_text=output_folder,
            size=(10, 1),
            border_width=2,
            key=" folder ")],
    ]
    outputFileSetting: list[list] = [
        [__sg.Text("Output File:")],
        [__sg.Input(
            default_text=str(output_file_prefix),
            size=(10, 1),
            border_width=2,
            key=" output "),
         __sg.Text(output_file_joint),
         __sg.Input(
            default_text=output_file_number[0] if isinstance(output_file_number, list) else str(output_file_number),
            size=(5, 1),
            border_width=2,
            key=" filenumber ") if output_file_number is not None else __sg.Text(),
         __sg.Text(F'{output_file_suffix}{output_file_extension}')]
    ]

    # gui layout
    layout: list[list] = [
        [__sg.Text(
            text="TDsampler",
            font=(font_family, title_font_size)),
         __sg.Frame(
            **commonFrameArgs,
            element_justification="right",
            layout=status,
            key=" statusfield ")],
        [__sg.HorizontalSeparator()],
        [__sg.Frame(
            **commonFrameArgs,
            element_justification="left",
            layout=startButton,
            key=" startbutton "),
         __sg.Text(
            text="",
            size=(maxTextLen + 7, None),
            key=" event ")],
        [__sg.Frame(
            **commonFrameArgs,
            element_justification=button_justification,
            layout=buttonField,
            key=" buttonfield ")],
        [__sg.HorizontalSeparator()],
        [__sg.Frame(
            **commonFrameArgs,
            element_justification="left",
            layout=outputFolderSetting,
            key=" outputfoldersetting "),
         __sg.VerticalSeparator(),
         __sg.Frame(
            **commonFrameArgs,
            element_justification="left",
            layout=outputFileSetting,
            key=" outputfilesetting ")]
    ]

    # MAIN WINDOW
    window = __sg.Window(
        title="TDsampler",
        layout=layout,
        resizable=True,
        finalize=True,
        return_keyboard_events=True,
        font=(font_family, font_size))

    # function that updates the gui components
    def updateWindow(keys: str | __Iterable, *args, **kwargs) -> __Callable:
        if isinstance(keys, str):
            keys = {keys}
        for key in keys:
            window[key].update(*args, **kwargs)

        return updateWindow

    # expand components to full width
    for key in {" statusfield ", " buttonfield "}:
        window[key].expand(expand_x=True)

    # hide the stop button and the save & reset button
    updateWindow({" SAVE & RESET ", " STOP "}, visible=False)

    # variables used in loop
    currentEvent: str = None  # current event
    event: str = None         # event
    values: str = None        # values of gui components
    startTime: float = 0.0    # time when the start button is pressed
    delay: float = 0.0        # time (seconds) from when the the start button is pressed to when the first button is pressed
    lapTime: float = 0.0      # time (seconds) from when the the start button is pressed to when the another button is pressed
    status: str = "initial"   # status of this application
    record: dict[str, list] = __initRecord(buttons)  # dict object that contains time info for each attribute

    # print status
    print("> Ready!\n")

    # start loop
    while True:
        # receive events and values of gui components
        [event, values] = window.read(timeout=10, timeout_key=" timeout ")

        # procedures corresponding to each event
        # when event is None
        if event is None:
            currentEvent and print("\n> Your task is cancelled.")
            break

        # when the start button is pressed
        elif event == " START " and status == "initial":
            # record timing
            startTime = __time.time()
            lapTime = 0.0

            # update gui components
            updateWindow(
                buttons, button_color=button_color_off)(
                " status ", "Recording ⏺", text_color="#ff0000")(
                " START ", visible=False)(
                " STOP ", visible=True)

            # print status
            print("[START]")

            # update variables
            currentEvent = " START "
            status = "started"

        # when the buttons (except start) are pressed
        elif event in buttons and event != currentEvent and status == "started":
            # record timing
            lapTime = round(__time.time() - startTime, 4)
            record[event].append(lapTime)

            # if the pressed button is the first button, set the delay time
            if currentEvent == " START ":
                delay = lapTime
            else:
                record[currentEvent].append(lapTime)
                updateWindow(currentEvent, button_color=button_color_off)

            # update gui components
            updateWindow(
                event, button_color=button_color_on)(
                " event ", F'NOW: {event}')

            # print the pressed button
            print(F' {event}')

            # update variables
            currentEvent = event

        # when the stop button is pressed
        elif event == " STOP " and status == "started":
            # record timing
            lapTime = round(__time.time() - startTime, 4)

            # if no buttons are pressed during the task, delay time will be equal to the end time
            if currentEvent == " START ":
                delay = lapTime
            else:
                record[currentEvent].append(lapTime)

            # add meta info to the record
            customMeta: dict = {}
            for key in custom_metadata:
                customMeta["#" + key.strip().upper().lstrip("#")] = custom_metadata[key]

            record |= customMeta
            record |= {"#ASSESSOR": [assessor_name],
                       "#DATE": [__datetime.datetime.now().strftime("%Y-%m-%d %H:%M")],
                       "#DURATION": [lapTime],
                       "#DELAY": [delay]}

            # update gui components
            updateWindow(
                buttons, button_color=button_color_disabled)(
                " status ", "Stopped ⏸", text_color="#0000ff")(
                " STOP ", visible=False)(
                " SAVE & RESET ", visible=True)(
                " event ", "")

            # print status
            print("[STOP]\n")

            # record score (option)
            if after_task_scoring:
                score = __scoringWindow(after_task_scoring)
                record |= {"#SCORE": [score],
                           "#SCORE_MIN": [after_task_scoring[0]],
                           "#SCORE_MAX": [after_task_scoring[1]]}

            # update valiables
            status = "stopped"

        # when the save & reset button is pressed
        elif event == " SAVE & RESET " and status == "stopped":
            # check if the file number is an integer
            fileNumber: int | str = values[" filenumber "]
            if isinstance(output_file_number, int) and output_file_increment:
                try:
                    fileNumber = int(fileNumber)
                except Exception:
                    __sg.popup(
                        "File number must be an integer\n",
                        title="Error",
                        font=(font_family, font_size))
                    continue

            # create file name
            outputFolderPath: str = F'{values[" folder "]}/'
            outputFileName: str = F'{values[" output "]}{output_file_joint}{fileNumber}' \
                                  F'{output_file_suffix}{output_file_extension}'

            # confirmation before saving file (notify when a file with the same name exists)
            beforeSaveConfirm: str = __sg.popup_yes_no(
                F'"{outputFileName}" already exists. Overwrite it?\n'
                if __os.path.isfile(F'{outputFolderPath}{outputFileName}')
                else F'Your record will be saved as "{outputFileName}". OK?\n',
                title="Confirm",
                font=(font_family, font_size)
            )

            # when instructed not to overwrite
            if beforeSaveConfirm != "Yes":
                continue

            # create new directory if it doesn't exist
            __newDir(outputFolderPath)

            # start saving
            try:
                with open(F'{outputFolderPath}{outputFileName}', "w", encoding="UTF-8") as outputFile:
                    outputFile.write(__dict2csv(record, comments))
            except Exception as e:
                __sg.popup(
                    "Failed to save your record\n",
                    title="Error",
                    font=(font_family, font_size))
                print(e)
                continue

            # print result
            __pprint.pprint(record)
            print()

            # notify completion
            __sg.popup(
                F'Your record has been saved in "{outputFileName}"\n',
                title="Success",
                font=(font_family, font_size))

            # increment or change file number
            if isinstance(output_file_number, list) and len(output_file_number) > 1:
                del output_file_number[0]
                updateWindow(" filenumber ", str(output_file_number[0]))

            elif isinstance(output_file_number, int) and output_file_increment:
                updateWindow(" filenumber ", str(fileNumber + 1))

            # restore gui components to initial state
            updateWindow(
                " SAVE & RESET ", visible=False)(
                " START ", visible=True, button_color=button_color_start)(
                " status ", "Not Started ⏹", text_color="#000000")

            # restore variables to initial state
            record = __initRecord(buttons)
            currentEvent = None
            status = "initial"

    # close window if the loop is broken
    print("> Bye!\n")
    window.close()


# window for product rating
def __scoringWindow(score_range: tuple[int]) -> int:

    score: int = (score_range[0] + score_range[1]) / 2

    layout: list[list] = [
        [__sg.Text("Please rate this product using the slider below!")],
        [__sg.HorizontalSeparator()],
        [__sg.Text()],
        [__sg.Slider(
            range=score_range,
            default_value=int(score),
            expand_x=True,
            orientation="horizontal",
            background_color=button_color_off,
            text_color=button_text_color,
            pad=(button_padding, button_padding),
            key=" score_slider "
        )],
        [__sg.Text()],
        [__sg.Submit(
            auto_size_button=True,
            focus=False,
            bind_return_key=False,
            button_text="OK",
            button_color=(button_text_color, button_color_off),
            size=(10, None),
            pad=((button_padding, button_padding), (0, button_padding * 2)),
            key=" score_submit ")
         ]
    ]

    window = __sg.Window(
        title="Scoring",
        layout=layout,
        resizable=True,
        finalize=True,
        return_keyboard_events=True,
        modal=True,
        disable_close=True,
        element_justification="center",
        font=(font_family, font_size))

    while True:
        [event, values] = window.read(timeout=10, timeout_key=" timeout ")

        if event is None:
            break
        if event == " score_submit ":
            score = int(values[" score_slider "])
            break

    window.close()

    return score


# FUNTCIONS
# make new directory if not exist
def __newDir(dirPath: str) -> None:
    if not __os.path.exists(dirPath):
        try:
            __os.mkdir(dirPath)
            print(F'{dirPath} was created successfully')
        except Exception as e:
            print(e)


# make new file if not exist
def __newFile(filePath: str, content: str) -> None:
    if not __os.path.isfile(filePath):
        try:
            with open(filePath, mode="w", encoding="UTF-8") as f:
                f.write(content)
                print(F'{filePath} was created successfully')
        except Exception as e:
            print(e)


# init record
def __initRecord(buttons: tuple[str]) -> dict[str, list]:
    return {"#DURATION": None, "#DELAY": None, **{elem: [] for elem in buttons}}


# dict to csv
def __dict2csv(dic: dict[str, list], comments: str | __Iterable[str] = None, delimiter: str = "\t") -> str:
    sortedDic = sorted(dic.items(), key=lambda x: x[0])
    result = "\n".join(F'{item[0]}{delimiter}{delimiter.join(map(str, item[1]))}'
                       for item in sortedDic) + "\n"

    if comments:
        if isinstance(comments, str):
            result += "\n: " + "\n: ".join(comments.replace("\r", "").split("\n")) + "\n"
        else:
            result += "\n: " + "\n: ".join(map(str, comments)) + "\n"

    return result


# create button labels from attribute.txt format string
def __attributetxt2list(string: str, shuffle: bool) -> tuple[str]:
    b = [*dict.fromkeys(elem for elem in map(lambda x: str.strip(x.split(":")[0]),
                                             dict.fromkeys(string.replace("\r", "\n").replace("\t", " ").split("\n")))
                        if len(elem) and elem[0] not in {"#", ":"})]
    shuffle and __random.shuffle(b)
    return b


if __name__ == "__main__":
    run()
